/* ##### Unstructured authentication tries ####
 * Simulate various authentications with random
 * - UID
 * - Tag challenge
 * - Encrypted reader challenge
 * - Encrypted reader answer
 * - Encrypted parity bits
 * 
 * A response from the tag is only obtained when all 8 parity bits are right.
 * This happens approx. all 256 tries. 
 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <bitset>
#include "mifaretag.h"
#include "uint48.h"

using namespace std;

int main(int argc, char **argv)
{
	srand((unsigned int)time(0));
	cout << "#### MIFARE CLASSIC simulation #####" << endl;
	
	// Initialize tag:
	MifareTag *mTag = new MifareTag(8);
	unsigned int tagUID = mTag->getUID();
	unsigned int sectorIndex = 0;
	unsigned int tagChallenge;
	
	// Open log file:
	ofstream logfile;
	logfile.open("successfulParitybits.log",ios::out);
	if(!logfile.is_open()) {
		cout << "Log file for successful parity attacks couldn't be opened!" << endl;
		return 1;
	}
	logfile << "ParityBits\tTagResponse\tTagUID\tTagChallenge\tReaderChallenge\tReaderAnswer\tSector" << endl;
	
	// Try several authentications:
	unsigned int tries = 0;
	unsigned int tagAnswers = 0;
	while(tries < 10000) {
		mTag->resetTag();
		tagChallenge = mTag->getChallenge(sectorIndex);
		
		unsigned int readerChallenge = rand();
		unsigned int readerAnswer = rand();
		unsigned int parityBits = rand() % 0x100;
		
		unsigned int tagAnswer = mTag->answerChallenge(readerChallenge,readerAnswer,parityBits);
		if(tagAnswer != 0xffffabcd) { // If the tag responds
			
			// Output parameters for this round, where the tag responded
			tagAnswers++;
			cout << std::dec << tries << ": Tag responded: " << std::hex << tagAnswer << endl;
			cout << std::dec << tagAnswers << ": " << std::hex << tagChallenge << " - " << readerChallenge << " - " << readerAnswer << " - " << parityBits << endl;
			logfile << std::hex << parityBits << "\t" << tagAnswer << "\t" << tagUID << "\t" << tagChallenge << "\t" << readerChallenge << "\t" << readerAnswer << "\t" << sectorIndex << endl;
		}
		for(int t=0;t<(rand() % 53);t++) {
			mTag->tick();
		}
		tries++;
	}
	cout << endl;
	
	logfile.close();
	cout << "Communication finished!" << endl;
	
	return 0;
}
