#include <iostream>
#include <fstream>
#include <string.h>
#include "../CRYPTO1/uint48.h"
#include "../CRYPTO1/lfsr.h"
#include "../CRYPTO1/helpfunctions.h"

using namespace std;

/* Convert a string containing a hexadecimal number to uint48 */
uint48 getKeyFromString(char *keyCharA)
{
	unsigned int length = strlen(keyCharA);
	unsigned int offset = 14-length;
	uint48 result;
	if(length < 3) {
		return result;
	}
	keyCharA[1] = '0';
	char high4[8] = { '0','0','0','0',keyCharA[2-offset], keyCharA[3-offset], keyCharA[4-offset], keyCharA[5-offset] };
	char low8[8] = { keyCharA[6-offset], keyCharA[7-offset], keyCharA[8-offset], keyCharA[9-offset], keyCharA[10-offset], keyCharA[11-offset], keyCharA[12-offset], keyCharA[13-offset] };
	unsigned int high = char2int(high4,8)[0];
	unsigned int low = char2int(low8,8)[0];
	result.setHighInt(high);
	result.setLowInt(low);
	return result;
	
}

/* Rollback the LFSR */
uint48 recoverKey(uint48 currentState, unsigned int encryptedReaderChallenge, unsigned int tagChallenge, unsigned int tagUID)
{
	LFSR *mLFSR = new LFSR();
	mLFSR->initialize(currentState);
	unsigned int ticks = 64;
	while(true)
	{
		currentState = mLFSR->getState();
		if(ticks > 64) {
			mLFSR->rollback(currentState);
		}
		else {
			if(ticks >= 32) {
				mLFSR->rollbackNRencrypted(currentState,encryptedReaderChallenge,ticks);
			}
			else {
				mLFSR->rollbackUID(currentState,tagUID,tagChallenge,ticks);
			}
		}
		if(ticks == 0) {
			break;
		}
		ticks--;
	}
	delete mLFSR;
	return currentState;	
}



int main(int argc, char **argv)
{
	// Input validation
	if(argc < 3) {
		cout << "Usage: CRYPTO1bruteForce <communication log> <states file>" << endl;
		return 1;
	}
	
	// Open files
	ifstream logfile;
	logfile.open(argv[1],ios::in);
	if(!logfile.is_open()) {
		cout << "The communication log file couldn't be opened!" << endl;
		return 2;
	}
	ofstream candidateFile;
	candidateFile.open("candidates.txt",ios::out);
	if(!candidateFile.is_open()) {
		cout << "The candidates file couldn't be opened!" << endl;
		return 2;
	}
	ifstream keyfile;
	keyfile.open(argv[2],ios::in);
	if(!keyfile.is_open()) {
		cout << "The state file couldn't be opened!" << endl;
		return 2;
	}
	
	cout << "###### Brute forcing CRYPTO 1 #######" << endl;
	
	// Definitions:
	string *headerNames = new string[7];
	unsigned int commData[64][7];
	uint48 LFSRStates[436];
	LFSR *mLFSR = new LFSR();
	unsigned int stateIndex = 0;
	char keyCharA[32];
	
	// Load online communication data from files
	logfile >> headerNames[0] >> headerNames[1] >> headerNames[2] >> headerNames[3] >> headerNames[4] >> headerNames[5] >> headerNames[6];
	unsigned int rows = 0;
	while(logfile.good() && !logfile.eof()) {
		logfile >> std::hex >> commData[rows][0] >> commData[rows][1] >> commData[rows][2] >> commData[rows][3] >> commData[rows][4] >> commData[rows][5] >> commData[rows][6];
		rows++;
	}
	if(rows == 0) {
		cout << "No data in communication log!" << endl;
		return 3;
	}	
	while(keyfile.good() && !keyfile.eof() && stateIndex < 436) {
		keyfile.getline(keyCharA,32);
		LFSRStates[stateIndex] = getKeyFromString(keyCharA);
		cout << "Key from file: " << keyCharA  << " -> " << std::hex << LFSRStates[stateIndex].getHighInt() << " " << LFSRStates[stateIndex].getLowInt() << endl;
		stateIndex++;
	}	
	
	// For all secrets:
	uint48 k, LFSRstate;
	
	unsigned int commDataRow = 0;
	unsigned int encryptedParityBits = commData[commDataRow][0];
	unsigned int encryptedErrorCode = commData[commDataRow][1]; // The output of the tag
	unsigned int expectedErrorCode = 0x5;
	unsigned int decodedErrorCode; // Used to test whether a specific key is valid
	unsigned int tagUID = commData[commDataRow][2];
	unsigned int tagChallenge = commData[commDataRow][3];
	unsigned int encryptedReaderChallenge = commData[commDataRow][4];
	unsigned int encryptedReaderAnswer = commData[commDataRow][5];
	unsigned int b_t;
	
	for(stateIndex=0;stateIndex<436;stateIndex++) {
		cout << " New state " << stateIndex << ": " << LFSRStates[stateIndex].getHighInt() << " " << LFSRStates[stateIndex].getLowInt() << endl;
		for(unsigned evenBits=0;evenBits<0x01000000;) {
			k = LFSRStates[stateIndex];
			for(int b=0;b<24;b++) {
				bool bit = (evenBits >> b) & 1;
				k.setBit(bit,2*b);
			}
			mLFSR->initialize(k);
			for(int t=64;t<96;t++) {
				mLFSR->stepNext();
			}
			
			encryptedParityBits = commData[commDataRow][0];
			encryptedErrorCode = commData[commDataRow][1];
			tagUID = commData[commDataRow][2];
			tagChallenge = commData[commDataRow][3];
			encryptedReaderChallenge = commData[commDataRow][4];
			encryptedReaderAnswer = commData[commDataRow][5];
			
			decodedErrorCode = 0;
			b_t = 0;
			for(int t=96;t<=99;t++) {
				b_t |= (mLFSR->getKeystream() << (99-t));
				mLFSR->stepNext();
			}
			decodedErrorCode = (encryptedErrorCode ^ b_t) & 0xf;
			
			if(expectedErrorCode == decodedErrorCode) {
				if(commDataRow == 0) {
					//uint48 keyCandidate = recoverKey()
					//cout << "Key candidate: " << std::hex << k.getHighInt() << k.getLowInt() << endl;
					//candidateFile << std::hex << k.getHighInt() << k.getLowInt() << endl;
				}
				if(commDataRow >= rows-2 || commDataRow > 6) {
					uint48 sectorKey = recoverKey(k,encryptedReaderChallenge,tagChallenge,tagUID);
					//candidateFile << "###  KEY: " << std::hex << k.getHighInt() << k.getLowInt();
					candidateFile << "0x" << std::hex << sectorKey.getHighInt() << sectorKey.getLowInt();
					candidateFile << "\t" << std::hex << LFSRStates[stateIndex].getHighInt() << LFSRStates[stateIndex].getLowInt();
					
					commDataRow = 0;
					evenBits++;
					continue;
				}
				commDataRow++;
			}
			else {
				if(commDataRow != 0) {
					//cout << "Candidate discarded! Passed " << commDataRow << " tests" << endl;
				}
				commDataRow = 0;
				evenBits++;
				
			}
		}
		
		
	}
	candidateFile.close();
	cout << "##### Brute force finished! ####" << endl;
	
	return 0;
}
