#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <bitset>
#include "mifaretag.h"
#include "uint48.h"

using namespace std;

unsigned int getParityBit(unsigned int responseParity, unsigned int currentParity, unsigned int keepFirstBits) {
	unsigned int parity = ((responseParity & 0xff) >> (8-keepFirstBits)) << (8-keepFirstBits);
	unsigned int increment;
	increment = (currentParity+1) % (1 << (8-keepFirstBits));
	return parity | increment;
}

int main(int argc, char **argv)
{
	/*LFSR *mLFSR = new LFSR();
	uint48 key; 
	uint48 variKey;
	unsigned int highInt = 0;
	unsigned int lowInt = 0;
	
	int i=0;
	srand((unsigned)time(0));
	key.setHighInt(rand() & 0xffff); key.setLowInt(rand());
	while(i < 100) {
		if(i % 2 == 0) {
			highInt = (key.getHighInt() & 0x00ff) | ((rand() % 0x100) << 8);
			lowInt = (key.getLowInt() & 0xffffffff) ;//| ((rand() % 0x1) << 20);
			key.setHighInt(highInt);
			key.setLowInt(lowInt);
		}
		if(i % 2 != 0) {
			highInt = key.getHighInt() ^ 0x00008000;
			variKey.setHighInt(highInt);
		}
		else {
			cout << endl;
			variKey.setHighInt(key.getHighInt());
			variKey.setLowInt(key.getLowInt());
		}
		mLFSR->initialize(variKey);
		
		bool keystream = mLFSR->getKeystream();
		cout << std::hex << std::setw(4) << std::setfill('0') << variKey.getHighInt() << variKey.getLowInt() << ": " << keystream << endl;
		i++;
	}
	cout << endl;
	return 0;*/
	
	/*int i=0;
	srand((unsigned)time(0));
	const int size = 5;
	bool input[5] = { 0, 0, 0, 0, 0 };
	while(i < 100) {
		if(i % 2 == 0)  {
			cout << endl;
			for(int j=0;j<size;j++) {
				if(j >= 0) input[j] = rand() % 2;
				cout << input[j] << " ";
			}
			cout << " - " << mLFSR->fc(input) << endl;
		}
		else {
			input[size-1] = (input[size-1]) ? false : true;
			for(int j=0;j<size;j++) {
				cout << input[j] << " ";
			}
			cout << " - " << mLFSR->fc(input) << endl;
		}
		i++;
	}
	cout << endl;
	return 0;
	*/
	
	srand((unsigned int)time(0));
	cout << "#### MIFARE CLASSIC simulation #####" << endl;
	
	// Initialize tag:
	MifareTag *mTag = new MifareTag(8);
	unsigned int tagUID = mTag->getUID();
	unsigned int sectorIndex = 0;
	unsigned int tagChallenge;
	
	// Open log file:
	ofstream logfile;
	logfile.open("successfulParitybits.log",ios::out);
	if(!logfile.is_open()) {
		cout << "Log file for successful parity attacks couldn't be opened!" << endl;
		return 1;
	}
	logfile << "ParityBits\tTagResponse\tTagUID\tTagChallenge\tReaderChallenge\tReaderAnswer\tSector" << endl;
	
	unsigned int authentifications = 0;
	unsigned int readerChallenge = rand();
	unsigned int readerAnswer = rand();
	unsigned int parityBits = 0;
	unsigned int challengeParityBits = 0x100;
	unsigned int variReaderChallenge;
	unsigned int tagAnswer;
	unsigned int missingTicks;
	bool Fj[4];
	unsigned int FjTestByte = 0;
	unsigned int bytesTested = 0;
	unsigned int moveTagChallenge = 0;
	bool testingNewReaderNonce = true;
	while(true) {
		moveTagChallenge = 0;
		mTag->resetTag();
		tagChallenge = mTag->getChallenge(sectorIndex);
		
		// Finding right parity:
		parityBits = getParityBit(challengeParityBits,parityBits,FjTestByte);
		if(testingNewReaderNonce) {
			variReaderChallenge = readerChallenge;
			if(authentifications % 0x10 == 0) cout << std::hex << parityBits << " - " << variReaderChallenge << " - " << tagChallenge <<  " - " << bytesTested << " - " << FjTestByte << endl;
		}
		else {
			variReaderChallenge = readerChallenge ^ (1 << (8*(3-FjTestByte)));
			if(authentifications % 0x10 == 0) cout << std::hex << parityBits << " - " << variReaderChallenge << " - " << tagChallenge <<  " - " << bytesTested << " - " << FjTestByte << endl;
		}
		tagAnswer = mTag->answerChallenge(variReaderChallenge,readerAnswer,parityBits);
		authentifications++;
		if(tagAnswer != 0xffffabcd) {
			bool b = true;
			if(!testingNewReaderNonce) {
				b = ((challengeParityBits ^ parityBits) >> (7-FjTestByte));
				if(!b) {
					Fj[FjTestByte] = true;
					bytesTested = 0;
				}
				else {
					Fj[FjTestByte] = false;
				}
				FjTestByte++;
			}
			else {
				challengeParityBits = parityBits;
				
				parityBits = 0;
			}
			
			
			cout << testingNewReaderNonce << " Tag responded: " << std::hex << tagAnswer << endl;
			cout << std::hex << tagChallenge << " - " << readerChallenge << " - " << readerAnswer << " - " << parityBits << " - " << challengeParityBits << " - " << b << endl;
			testingNewReaderNonce = false;
		}
		if(FjTestByte == 4 && Fj[3]) { // Stop if all 4 bits have property Fj
			break;
		}
		if(FjTestByte > 0 && !Fj[FjTestByte-1] ) { // Change jth byte if it has not property Fj 
			cout << "Going back" << endl;
			FjTestByte--;
			if((FjTestByte == 0 && bytesTested >= 0x14) 
				|| (FjTestByte > 0 && bytesTested >= 0x14) ) {
				if(FjTestByte > 0) {
					FjTestByte--;
				}
				else {
					moveTagChallenge = 1;
				}
				bytesTested = 0;
			}
				unsigned char byte = (readerChallenge >> 8*(3-FjTestByte)) & 0xff;
				byte = (byte-1) % 0x100;
				unsigned int mask = 0xffffff00;
				switch(FjTestByte) {
					case 2:
						mask = 0xffff00ff;
						break;
					case 1:
						mask = 0xff00ffff;
						break;
					case 0:
						mask = 0x00ffffff;
						break;
				}
				readerChallenge = (byte << 8*(3-FjTestByte)) | (readerChallenge & mask);
				bytesTested++;
				testingNewReaderNonce = true;
				
			parityBits = 0;
		}
		
		// Keep nT constant:
		missingTicks = 65534+moveTagChallenge-(mTag->getTicks() % 65535);
		for(unsigned int t=0;t<missingTicks;t++) {
			mTag->tick();
		}
	}
	double timeElapsed = authentifications*0.318;
	cout << endl << "After " << std::dec << authentifications << " auths (" << setprecision(4) <<  timeElapsed << "s): " << std::hex << readerChallenge << " has Fj property: " << endl;
	for(int i=0;i<4;i++) {
		cout << Fj[i] << " - ";
	}
	cout << endl << endl;
	cout << "#### Interesting combination: #####" << endl;
	cout << "TagChallenge: " << std::hex << tagChallenge << endl;
	cout << "ReaderChallenge: " << std::hex << readerChallenge << endl;
	cout << "ReaderAnswer: " << std::hex << readerAnswer << endl;
	cout << "ParityBits: " << std::hex << challengeParityBits << endl;
	cout << "###########################" << endl;
	cout << endl;
	logfile << std::hex << challengeParityBits << "\t" << 0 << "\t" << tagUID << "\t" << tagChallenge << "\t" << readerChallenge << "\t" << readerAnswer << "\t" << sectorIndex << endl;

	
	logfile.close();
	cout << "Communication finished!" << endl;
	
	return 0;
}
