#include <iostream>
#include <fstream>
#include <string.h>
#include "../CRYPTO1/uint48.h"
#include "../CRYPTO1/lfsr.h"
#include "../CRYPTO1/helpfunctions.h"

using namespace std;

uint48 getKeyFromString(char *keyCharA)
{
	unsigned int length = strlen(keyCharA);
	unsigned int offset = 14-length;
	uint48 result;
	if(length < 3) {
		return result;
	}
	keyCharA[1] = '0';
	char high4[8] = { '0','0','0','0',keyCharA[2-offset], keyCharA[3-offset], keyCharA[4-offset], keyCharA[5-offset] };
	char low8[8] = { keyCharA[6-offset], keyCharA[7-offset], keyCharA[8-offset], keyCharA[9-offset], keyCharA[10-offset], keyCharA[11-offset], keyCharA[12-offset], keyCharA[13-offset] };
	unsigned int high = char2int(high4,8)[0];
	unsigned int low = char2int(low8,8)[0];
	result.setHighInt(high);
	result.setLowInt(low);
	return result;
	
}

int main(int argc, char **argv)
{
	if(argc < 2) {
		cout << "Usage: CRYPTO1bruteForce <communication log>" << endl;
		return 1;
	}
	ifstream logfile;
	logfile.open(argv[1],ios::in);
	if(!logfile.is_open()) {
		cout << "The communication log file couldn't be opened!" << endl;
		return 2;
	}
	ofstream candidateFile;
	candidateFile.open("candidates.txt",ios::out);
	if(!candidateFile.is_open()) {
		cout << "The candidates file couldn't be opened!" << endl;
		return 2;
	}
	bool useKeyfile = false;
	ifstream keyfile;
	if(argc >= 3) {
		keyfile.open(argv[2],ios::in);
		if(keyfile.is_open()) {
			useKeyfile = true;
		}
	}
	cout << "###### Brute forcing CRYPTO 1 #######" << endl;
	
	// Load communication log:
	string *headerNames = new string[7];
	unsigned int commData[64][7];
	logfile >> headerNames[0] >> headerNames[1] >> headerNames[2] >> headerNames[3] >> headerNames[4] >> headerNames[5] >> headerNames[6];
	//cout << headerNames[0] << "\t" << headerNames[1] << "\t" << headerNames[2] << "\t" << headerNames[3] << "\t" << headerNames[4] << "\t" << headerNames[5] << "\t" << headerNames[6] << endl;
	unsigned int rows = 0;
	while(logfile.good() && !logfile.eof()) {
		logfile >> std::hex >> commData[rows][0] >> commData[rows][1] >> commData[rows][2] >> commData[rows][3] >> commData[rows][4] >> commData[rows][5] >> commData[rows][6];
		//cout << std::hex << commData[rows][0] << "\t" << commData[rows][1] << "\t" << commData[rows][2] << "\t" << commData[rows][3] << "\t" << commData[rows][4] << "\t" << commData[rows][5] << "\t" << commData[rows][6] << endl;
		rows++;
	}
	if(rows == 0) {
		cout << "No data in communication log!" << endl;
		return 3;
	}
	
	LFSR *mLFSR = new LFSR();
	
	// For all secrets:
	uint48 kStart, k, LFSRState;
	
	unsigned int decodedParityBits;
	unsigned int encryptedErrorCode;
	unsigned int expectedErrorCode = 0x5;
	unsigned int decodedReaderChallenge;
	unsigned int decodedReaderAnswer;
	bool b_t;
	unsigned int parityBitCounter;
	unsigned int expectedParityBits;
	
	kStart.setHighInt(0);
	kStart.setLowInt(0);
	unsigned int commDataRow = 0;
	for(uint48 k=kStart;k.getHighInt()<=0x0000ffff;) {
		if(useKeyfile && commDataRow == 0) {
			char keyCharA[32];
			if(keyfile.good() && !keyfile.eof()) {
				keyfile.getline(keyCharA,32);
			}
			else {
				break;
			}
			k = getKeyFromString(keyCharA);
			cout << "Key from file: " << keyCharA  << " -> " << std::hex << k.getHighInt() << " " << k.getLowInt() << endl;
		}
		/* Initialize the LFSR, i.e. input the sector key */
		mLFSR->initialize(k);
		
		/* Read data from communication log */
		decodedParityBits = commData[commDataRow][0];
		encryptedErrorCode = commData[commDataRow][1];
		expectedErrorCode = 0x5;
		decodedReaderChallenge = commData[commDataRow][4];
		decodedReaderAnswer = commData[commDataRow][5];
		parityBitCounter = 0;
		expectedParityBits = 0;
		
		/* Shift in the data into the LFSR */
		/* Shift in the tags UID and the tag challenge */
		for(int t=0;t<32;t++) {
			mLFSR->stepNext(commData[commDataRow][2],commData[commDataRow][3],t);
		}
		/* Decrypt and shift in the reader challenge 
		 * and calculate decrypted parity bits
		 */
		for(int t=32;t<=64;t++) {
			b_t = mLFSR->getKeystream();
			if(t<64) {
				decodedReaderChallenge ^= (b_t << (63-t));
				mLFSR->stepNext(decodedReaderChallenge,t);
			}
			if(t != 32 && t % 8 == 0) {
				unsigned char byte = (decodedReaderChallenge >> (8*(3-parityBitCounter))) & 0xff;
				bool parityByte = calculateParityBit(byte);
				expectedParityBits |= parityByte << (7-parityBitCounter);
				decodedParityBits ^= (b_t << (7-parityBitCounter));
				parityBitCounter++;
			}
		}
		/* Decrypt the reader answer 
		 * and calculate decrypted parity bits
		 */
		for(int t=64;t<=96;t++) {
			b_t = mLFSR->getKeystream();
			if(t < 96) {
				decodedReaderAnswer ^= (b_t << (95-t));
				mLFSR->stepNext();
			}
			if(t != 64 && t % 8 == 0) {
				unsigned char byte = (decodedReaderAnswer >> (8*(7-parityBitCounter))) & 0xff;
				bool parityByte = calculateParityBit(byte);
				expectedParityBits |= parityByte << (7-parityBitCounter);
				decodedParityBits ^= (b_t << (7-parityBitCounter));
				parityBitCounter++;
			}
		}
		/* Decrypt the tag error code */
		for(int t=96;t<100;t++) {
			b_t = mLFSR->getKeystream();
			expectedErrorCode ^= (b_t << (99-t));
			mLFSR->stepNext();
		}
		
		/* Check if the currently tested key reproduces
		 * the observed parity bits and encrypted error code
		 */
		if(expectedParityBits == decodedParityBits && expectedErrorCode == encryptedErrorCode) {
			if(commDataRow == 0) {
				cout << "Key candidate: " << std::hex << k.getHighInt() << k.getLowInt() << endl;
				
				/* Output candidate to file. Uncomment if needed:*/
				//candidateFile << std::hex << k.getHighInt() << k.getLowInt() << endl;
			}
			if(commDataRow >= rows-2 || commDataRow > 6) {
				cout << "###  KEY: " << std::hex << k.getHighInt() << k.getLowInt() << endl;
				candidateFile << "###  KEY: " << std::hex << k.getHighInt() << k.getLowInt() << endl;
				break;
			}
			commDataRow++;
		}
		else {
			if(commDataRow != 0) {
				cout << "Candidate discarded! Passed " << commDataRow << " tests" << endl;
			}
			commDataRow = 0;
			k++;
			
		}
		
		/* Output progress of brute force search */
		if(k.getLowInt() % 200000 == 0) {
			cout << std::dec << k.getHighInt() << k.getLowInt() << " keys tested" << endl;
		}
	}
	candidateFile.close();
	cout << "##### Brute force finished! ####" << endl;
	
	return 0;
}
