/* ##### 48 bit unsigned integer #####
 * The class emulates an unsigned integer with 48 bits.
 * The implementation is minimal to match the CRYPTO1 project.
 * The data are stored in two 32 bit unsigned integers
 */

#include "uint48.h"

/* ###### Constructors ######## */
uint48::uint48()
{
	uint48(0, 0);
}

uint48::uint48(unsigned int _high, unsigned int _low)
{
	high = _high;
	low = _low;
}

uint48::~uint48()
{
}

/* ###### Operators ######## */
uint48 uint48::operator <<(int shift)
{
	if(shift < 0) {
		*this >> -shift;
	}
	else
	{
		high = (high << shift) | (low >> (32-shift));
		high = high & 0x0000ffff;
		low = low << shift;
	}
	return *this;
}

uint48 uint48::operator >>(int shift)
{
	if(shift < 0) {
		*this << -shift;
	}
	low = (low >> shift) | (high << (32-shift));
	high = high >> shift;
	return *this;
}

uint48 uint48::operator ^(const uint48 &right)
{
	uint48 result;
	result.setHighInt(high ^ right.high);
	result.setLowInt(low^right.low);
	return result;
}

uint48 uint48::operator |(const uint48& right)
{
	uint48 result;
	result.setHighInt(high|right.high);
	result.setLowInt(low|right.low);
	return result;
}

uint48 uint48::operator &(const uint48& right)
{
	uint48 result;
	result.setHighInt(high&right.high);
	result.setLowInt(low&right.low);
	return result;
}


uint48 uint48::operator &(const unsigned int right)
{
	uint48 result;
	result.setHighInt(high);
	result.setLowInt(low&right);
	return result;
}

uint48 uint48::operator ^(const unsigned int right)
{
	uint48 result;
	result.setHighInt(high);
	result.setLowInt(low^right);
	return result;
}

uint48 uint48::operator |(const unsigned int right)
{
	uint48 result;
	result.setHighInt(high);
	result.setLowInt(low|right);
	return result;
}

uint48 uint48::operator ~()
{
	uint48 result;
	result.setHighInt(~high & 0x0000ffff);
	result.setLowInt(~low);
	return result;
}


uint48 uint48::operator =(const uint48& other)
{
	if(this == &other) {
		return *this;
	}
	high = other.high;
	low = other.low;
	return *this;
}

uint48 uint48::operator =(const unsigned int other)
{
	high = 0;
	low = other;
	return *this;
}

uint48 uint48::operator ++(const int notused)
{
	if(low == 0xffffffff) {
		low = 0;
		high++;
	}
	else {
		low++;
	}
	return *this;
}


/* ###### Getters and setters ###### */
const unsigned int uint48::getHighInt()
{
	return high;
}

const unsigned int uint48::getLowInt()
{
	return low;
}

bool uint48::getBit(unsigned int position)
{
	if(position >= 48) {
		return false;
	}
	if(position >= 32) {
		return (high >> (position-32)) & 0x1; 
	}
	else {
		return (low >> position) & 0x1;
	}
}

void uint48::setHighInt(unsigned int _high)
{
	high = _high & 0x0000ffff;
}

void uint48::setLowInt(unsigned int _low)
{
	low = _low;
}

void uint48::setBit(bool bit, unsigned int position)
{
	if(position >= 48) {
		return;
	}
	if(position >= 32) {
		high |= (bit << (position-32));
	}
	else {
		low |= (bit << position);
	}
}

